/*******************************************************
 ***
 ***      Arithmetic Interpreter        Version 2.2
 ***
 ***               Objective-C Version
 ***
 *******************************************************/

/*

Cette portion de code permet de reconnaitre des expressions mathematiques
rentrees sous forme de chaines de caracteres et ensuite d'evaluer ces ex-
pressions.

Les fonctions reconnues sont:
+ - * / ^
exp ln log abs sin cos tan asin acos atan sinh (ou sh) cosh (ou ch)
tanh (ou th)
L'interpreteur reconnait les nombres reels,entiers positifs ou negatifs.
L'interpreteur reconnait les variables x et y.

L'interpreteur sait evaluer des expressions avec redondance de parentheses.
Exemple: (1+(2)) donnera 3.

Les fonctions unaires (sin,cos,abs,....), si elles n'admettent qu'un seul 
argument (Exemple sin(x)) peuvent etre ecrite aussi sans parentheses (Exemple:
sinx).   ATTENTION: ecrire sinx+3 est equivalent a sin(x)+3.

Pour beneficier de cet interpreteur, il suffit d'inclure le fichier OCInterpreter.h
dans le projet et de faire appel aux fonctions:

OCNode * ConstructTree(char * string,short length,short * error);
-(float) ComputeTreewithXValue:(float) xvalue YValue:(float) yvalue error:(short *) error;


La fonction ConstructTree sert a construire l'arbre binaire representant l'expres-
sion saisie au clavier.Elle prend en arguments une chaine de caractere a interpreter
, le nombre de caracteres de cette chaine (signe '\0' non inclus) et un short * qui 
representera le type d'erreur rencontree.

IMPORTANT: les lettres doivent etre en minuscule et il ne doit pas y'avoir d'espace
sous peine de renvoi d'un code d'erreur 1,2 ou 3.

exemple: 3*x+1 est correct. 3*X+1  non      -> erreur de type 3. 
                            3* x+1 non plus -> erreur de type 3.

La fonction ComputeTreewithXValue prend en arguments les valeurs respectives des variables
x et y et un short qui representera le type d'erreur rencontree.

Statistiques:
	
Sur la fonction: 1/(x+0.2)*sin(x^2)*cos(x), on obtient ces resultats somme toute
corrects vu que c'est de l'Objective-C:

La fonction n'est interpretee qu'une fois et evaluee ensuite 200 000 fois.

135000 evaluations/seconde sur un PowerMac G3/266  Rhpasody DR2 (Non optimise donc).
162000 evaluations/seconde sur une SUN Ultra 5     Solaris
 		
Teste sur:

SUN Ultra 5:
SOLARIS                 gcc -o Interpreter OCInterpreter.m -lobjc -lm

PowerMac G3:
MacOS X Server (RDR2)   Faire un Project Tool sous Project Builder.
MacOS 8.x               CodeWarrior IDE (MW RunTime)

Historique:

2.2:
    * Version Objective-C disponible.
    * Amelioration du code de la version C++
      - Correction de bugs sur l'operator =
      - Optimisation du code
      - Reformatage pour plus de clarete.
    * Mise a jour du code de la version C.

2.1:
  * Version Java disponible:
    # les operateurs mathematiques disponibles sont ceux de Java donc certains manquent.

  * Correction d'un bug sur le calcul de puissance (Version C++).

2.0: 
  * Portage en C++ et debut en Java

Version C++:
	 
  * Meilleure gestion memoire lors de la construction de l'expression.
  * Acceleration de certaines operations.

Version Java:
	 
  * Premiere version. Alors si il y'a un bug, c'est fort possible.

1.3b: ajoute les fonctions suivantes: (NON DISTRIBUEE)
  * reconnaissance du symbole 
  * ecriture formatee d'expression (a ameliorer)

1.2: corrige les bugs suivants:
  * erreur sur l'interpretation de fonctions unaires imbriquees telles que
    ln(exp(x)).
  * la fonction puissance (pour les puissances entieres).
    ajoute:
  * la verification de la chaine entree (voir code d'erreur)
  * la verification de l'existence de l'evaluation (voir code d'erreur)
     

1.1: corrige un bug au niveau de l'interpretation des fonctions du type:

      exp(-(x+y))

1.0: premiere version


Ce code peut etre distribue et modifie librement.

*/


#if MACOSX
#import "OCInterpreter.h"
#else
#include "OCInterpreter.h"
#endif

#include <math.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

char * CopyPartialString(char * inString,short inStart,short inEnd) {
char *  tString;
short tLength=inEnd-inStart+1;

tString=(char *) calloc(tLength+1,sizeof(char));

if (tString!=NULL) {
  memmove(tString,inString+inStart,tLength);
  tString[tLength]='\0';
}

return tString;
}

float CalcValue(char * inString,short inIndex) {
short i=inIndex,j;
float tValue=0;
short signe=1;

/*if (chaine[inIndex]=='') return PI;*/
if (inString[inIndex]=='-') {
	signe=-1;
	i++;
}

while (inString[i]>47 && inString[i]<58) {
	tValue=tValue*10+(inString[i]-'0');
	i++;
}

if (inString[i]=='.') {
	i++;
	j=1;
	while (inString[i]>47 && inString[i]<58) {
		tValue=tValue+(inString[i]-'0')*exp(-j*2.30258509);
		i++;
		j++;
	}
}

return (signe*tValue);
}

void FindOperator(short * oper,short * esp,char * inString,short pos) {

switch(inString[pos]) {
	case 'a':
		switch(inString[pos+1]) {
			case 'b':
				*esp=3;
				*oper=10;
				break;
			case 'c':
				*esp=4;
				*oper=15;
				break;
			case 's':
				*esp=4;
				*oper=14;
				break;
			case 't':
				*esp=4;
				*oper=16;
				break;
		}
		break;
	case 'c':
		if (inString[pos+1]=='h') {
			*esp=2;
			*oper=18;
		}
		else if (inString[pos+1]=='o' && inString[pos+2]=='s') {
			if (inString[pos+3]=='h') {
				*esp=4;
				*oper=18;
			}
			else {
        *esp=3;
				*oper=12;
			}
		}
		break;
	case 'e':
		if (inString[pos+1]=='x' && inString[pos+2]=='p') {
			*esp=3;
         	*oper=6;
		}
		else *oper=-10;
		break;
	case 'l':
		if (inString[pos+1]=='n') {
			*esp=2;
			*oper=7;
		}
		else if (inString[pos+1]=='o' && inString[pos+2]=='g'){
         	*esp=3;
            *oper=8;
         }
         else *oper=-10;
         break;
	case 's':
		if (inString[pos+1]=='h') {
			*esp=2;
			*oper=17;
		}
		else if (inString[pos+1]=='q') {
			*esp=4;
			*oper=9;
		}
		else {
			if (inString[pos+3]=='h') {
				*esp=4;
				*oper=17;
			}
			else {
				*esp=3;
				*oper=11;
			}
		}
		break;
	case 't':
		if (inString[pos+1]=='h') {
			*esp=2;
			*oper=19;
		}
		else if (inString[pos+1]=='a' && inString[pos+2]=='n') {
			if (inString[pos+3]=='h') {
				*esp=4;
				*oper=19;
			}
			else {
				*esp=3;
				*oper=13;
			}
		}
		else *oper=-10;
		break;
	default:
		*oper=-10;
		break;
}
}

@implementation OCNode

-(id) init
{
[super init];

fg_=fd_=NULL;

return self;
}

-(id) initWithOperator:(short) Operator value:(float) value fg:(OCNode *) fg fd:(OCNode *) fd
{
[super init];

operator_=Operator;
value_=value;
fg_=fg;
fd_=fd;

return self;
}

-(id) setFg:(OCNode *) fg
{
fg_=fg;
return self;
}

-(id) setFd:(OCNode *) fd
{
fd_=fd;
return self;
}

-(float) ComputeTreewithXValue:(float) xvalue YValue:(float) yvalue error:(short *) error
{
float valueL;
float valueR;

if (operator_==0) return value_;
if (operator_==-1) return xvalue;
if (operator_==-2) return yvalue;

valueL=[fg_ ComputeTreewithXValue:xvalue YValue:yvalue error:error];

if (*error!=0) return 0;


if (fd_!=NULL) valueR=[fd_ ComputeTreewithXValue:xvalue YValue:yvalue error:error];
if (*error!=0) return 0;

switch(operator_) {
	case 1:								/* + */
		return (valueL+valueR);
	case 2:								/* - */
		return (valueL-valueR);
	case 3:								/* * */
		return (valueL*valueR);
	case 4:								/* - */
		if (valueR==0) {
			*error=1;
			return 0;
		}
		return (valueL/valueR);
	case 5:								/* ^ */
		if ((((long) valueR)-valueR)==0) return pow(valueL,valueR);
		else if (valueL<0) {
			*error=2;
			return 0;
		} 
		else return pow(valueL,valueR);
		break;
	case 6:								/* exp */
		return exp(valueL);
	case 7:								/* ln */
		if (valueL<=0) {
			if (valueL<0) *error=2;
			else *error=1;
			return 0;
		}
		return log(valueL);
	case 8:								/* log */
		if (valueL<=0) {
			if (valueL<0) *error=2;
			else *error=1;
			return 0;
		}
		return log10(valueL);
	case 9:								/* sqrt */
		if (valueL<0) {
			*error=2;
			return 0;
		}
		return sqrt(valueL);
	case 10:							/* abs */
		return fabs(valueL);
	case 11:
		return sin(valueL);			    /* sin */
	case 12:
		return cos(valueL);			    /* cos */
	case 13:
		return tan(valueL);			    /* tan */
	case 14:
		return asin(valueL);			/* asin */
	case 15:
		return acos(valueL);			/* acos */
	case 16:
		return atan(valueL);			/* atan */
	case 17:
		return sinh(valueL);			/* sinh */
	case 18:
		return cosh(valueL);			/* cosh */
	case 19:
		return tanh(valueL);			/* tanh */
	default:
		return 0;
}
}

-(void) WriteTree:(char *) inString
{
int parenthese=0;
char tString[20];

switch(operator_) {
	case -2:
		strcat(inString,"y");
		break;
	case -1:
		strcat(inString,"x");
		break;
	case 0:
		
		sprintf(tString,"%f",value_);
		strcat(inString,tString);
		break;
	case 1:
		[fg_ WriteTree:inString];
		strcat(inString,"+");
		[fd_ WriteTree:inString];
		break;
	case 2:
		if (fg_->operator_!=0 || fg_->value_!=0)
		  [fg_ WriteTree:inString];
		strcat(inString,"-");
		if (fd_->operator_==1 || fd_->operator_==2) {
			parenthese=1;
			strcat(inString,"(");
		}
		[fd_ WriteTree:inString];
		if (parenthese==1) {
			strcat(inString,")");
		}
		break;
	case 3:
		if (fg_->operator_==1 || fg_->operator_==2) {
			parenthese=1;
			strcat(inString,"(");
		}
		[fg_ WriteTree:inString];
		if (parenthese==1)
			strcat(inString,")");
		parenthese=0;
		strcat(inString,"*");
		if (fd_->operator_==1 || fd_->operator_==2) {
			parenthese=1;
			strcat(inString,"(");
		}
		[fd_ WriteTree:inString];
		if (parenthese==1)
			strcat(inString,")");
		break;
	case 4:
		if (fg_->operator_==1 || fg_->operator_==2) {
			parenthese=1;
			strcat(inString,"(");
		}
		[fg_ WriteTree:inString];
		if (parenthese==1)
			strcat(inString,")");
		parenthese=0;
		strcat(inString,"/");
		if (fd_->operator_>0 && fd_->operator_<5) {
			parenthese=1;
			strcat(inString,"(");
		}
		[fd_ WriteTree:inString];
		if (parenthese==1) 
			strcat(inString,")");
		break;
	case 5:
		if (fg_->operator_>0 && fg_->operator_<5) {
			parenthese=1;
			strcat(inString,"(");
		}
		[fg_ WriteTree:inString];
		if (parenthese==1)
			strcat(inString,")");
		parenthese=0;
		strcat(inString,"^");
		if (fd_->operator_>0 && fd_->operator_<5) {
			parenthese=1;
			strcat(inString,"(");
		}
		[fd_ WriteTree:inString];
		if (parenthese==1) 
			strcat(inString,")");
		break;
	case 6:
		strcat(inString,"exp(");
		[fg_ WriteTree:inString];
		strcat(inString,")");
		break;
	case 7:
		strcat(inString,"log(");
		[fg_ WriteTree:inString];
		strcat(inString,")");
		break;
	case 8:
		strcat(inString,"ln(");
		[fg_ WriteTree:inString];
		strcat(inString,")");
		break;
	case 9:
		strcat(inString,"sqrt(");
		[fg_ WriteTree:inString];
		strcat(inString,")");
		break;
	case 10:
		strcat(inString,"|");
		[fg_ WriteTree:inString];
		strcat(inString,"|");
		break;
	case 11:
		strcat(inString,"sin(");
		[fg_ WriteTree:inString];
		strcat(inString,")");
		break;
	case 12:
		strcat(inString,"cos(");
		[fg_ WriteTree:inString];
		strcat(inString,")");
		break;
	case 13:
		strcat(inString,"tan(");
		[fg_ WriteTree:inString];
		strcat(inString,")");
		break;
	case 14:
		strcat(inString,"asin(");
		[fg_ WriteTree:inString];
		strcat(inString,")");
		break;
	case 15:
		strcat(inString,"acos(");
		[fg_ WriteTree:inString];
		strcat(inString,")");
		break;
	case 16:
		strcat(inString,"atan(");
		[fg_ WriteTree:inString];
		strcat(inString,")");
		break;
	case 17:
		strcat(inString,"sinh(");
		[fg_ WriteTree:inString];
		strcat(inString,")");
		break;
	case 18:
		strcat(inString,"cosh(");
		[fg_ WriteTree:inString];
		strcat(inString,")");
		break;
	case 19:
		strcat(inString,"tanh(");
		[fg_ WriteTree:inString];
		strcat(inString,")");
		break;
}
}

@end

OCNode * ConstructTree(char * string,short length,short * error) {
short imbric,Bimbric;
short priorite,ope;
short position,positionv,i,j;
short opetemp=0;
short espa,espat;
short caspp=0;
char * tString;
OCNode * node;

/* Initialisation des variables */

if (length<=0) {
	*error=3;
	return NULL;
}

ope=0;
imbric=0;Bimbric=128;
priorite=6;
i=0;
positionv=position=0;

/* Mise en place des donnees sur le morceau de chaine */

while (i<length) {
	/*printf("%c",string[i]);*/
	if (((string[i]>47) && (string[i]<58)) || (string[i]=='')) {
		if (priorite>5) {
			priorite=5;
			positionv=i;
		}
		i++;
	}
	else {
	  if ((string[i]>96) && (string[i]<117)) {
		FindOperator(&opetemp,&espat,string,i);
		if (opetemp>=0) {
			if (imbric<Bimbric) {
				Bimbric=imbric;
				ope=opetemp;
				position=i;
				priorite=4;
				espa=espat;
			} 
			else if ((imbric==Bimbric) && (priorite >=4)) {
				ope=opetemp;
				position=i;
				priorite=4;
				espa=espat;
			}
			j=i+1;
			i+=espat;
			while(j<i) {
				/*printf("%c",string[j]);*/
				j++;
			}
		}
		else if (string[i]=='t') {
			if (priorite==6) ope=-1;
			i++;
		}
		else if (string[i]=='p') {
			if (priorite==6) ope=-2;
			i++;
		}
		else if (string[i]=='r') {
			if (priorite==6) ope=-2;
			i++;
		}
		else if (string[i]=='n') {
			if (priorite==6) ope=-1;
			i++;
		}
		else {
			*error=2; /* symbole non reconnu */
      return NULL;
    }
	}
	else {
		switch(string[i]) {
			case '(':
				imbric++;
				break;
			case ')':
				imbric--;
				break;
			case '+':
				if (imbric<Bimbric) {
					Bimbric=imbric;
					priorite=1;
					ope=1;
					position=i;
					caspp=0;
				}
				else if ((imbric==Bimbric) && (priorite >=1)) {
					priorite=1;
					ope=1;
					position=i;
					caspp=0;
				}
				break;
			case '-':
				if (imbric<Bimbric) {
					if ((i-1)<0) {
						if (priorite>5) {
							priorite=1;
							position=i;
							ope=2;
							Bimbric=imbric;
							caspp=1;
						}
					}
				 	else if (string[i-1]=='(') {
				 		if (priorite>1){
				 			priorite=1;
				 			position=i;
				 			Bimbric=imbric;
				 			caspp=1;
				 			ope=2;
				 		}
				 	}
					else {
						Bimbric=imbric;
						priorite=1;
						ope=2;
						position=i;
						caspp=0;
					}
				}
				else if ((imbric==Bimbric) && (priorite>=1)) {
					if ((i-1)<0) {
						if (priorite>5) {
							priorite=1;
							position=i;
							ope=2;
							caspp=1;
						}
					}
					else if (string[i-1]=='(') {
						if (priorite>1){
							priorite=1;
							position=i;
							caspp=1;
							ope=2;
						}
					}
					else {
						priorite=1;
						ope=2;
						position=i;
						caspp=0;
					}
				}
				break;
			case '*':
				if (imbric<Bimbric) {
					Bimbric=imbric;
					priorite=2;
					ope=3;
					position=i;
				}
				else if ((imbric==Bimbric) && (priorite>=2)) {
					priorite=2;
					ope=3;
					position=i;
				}
				break;
			case '/':
				if (imbric<Bimbric) {
					Bimbric=imbric;
					priorite=2;
					ope=4;
					position=i;
				}
				else if ((imbric==Bimbric) && (priorite>=2)) {
					priorite=2;
					ope=4;
					position=i;
				}
				break;
			case '^':
				if (imbric<Bimbric) {
					Bimbric=imbric;
					priorite=3;
					ope=5;
					position=i;
				}
				else if ((imbric==Bimbric) && (priorite>=3)) {
					priorite=3;
					ope=5;
					position=i;
				}
				break;
			case '.':
				break;
			case 'x':
				if (priorite==6) ope=-1;
				break;
			case 'y':
				if (priorite==6) ope=-2;
				
				break;
			default:
				*error=2; /* symbole non reconnu */
				return NULL;
			}
      i++;
		}
	}
}

if (imbric!=0) {
	*error=1; /* erreur de parenthesage */
	return NULL;
}

/* Traitement des donnees	

/printf("\nMeilleure Imbrication: %d Meilleure priorite: %d Operateur: %d",Bimbric,priorite,ope);
/printf(" Position: %d Positionv: %d\n",position,positionv);
/printf("caspp: %d\n",caspp);*/


if (priorite==6) {
	node = [[OCNode alloc] initWithOperator:ope value:0 fg:NULL fd:NULL];
    
  return node;
}
else if (caspp==1) {
	node = [[OCNode alloc] initWithOperator:2 value:0 fg:NULL fd:NULL];
	
	[node setFg:[[OCNode alloc] initWithOperator:2 value:0 fg:NULL fd:NULL]];
				    
	if ((length-position-1-Bimbric)==0) { /* argument absent */
		*error=3;
		return NULL;
	} 
	tString=CopyPartialString(string,(position+1),(length-1-Bimbric));
	
  [node setFd:ConstructTree(tString,(length-position-1-Bimbric),error)];
	
  free(tString);
	
  return node;
}
else if (priorite==5) {
	node = [[OCNode alloc] initWithOperator:0 value:CalcValue(string,positionv) fg:NULL fd:NULL];
	
  return node;
}
else if (ope>5) {
	node = [[OCNode alloc] initWithOperator:ope value:0 fg:NULL fd:NULL];

	
	if ((length-position-espa-Bimbric)==0) { /* argument absent */
		*error=3;
		return NULL;
	}
	
  tString=CopyPartialString(string,(position+espa),(length-1));
	
  [node setFg:ConstructTree(tString,(length-position-espa-Bimbric),error)];
	
  free(tString);
	
  return node;
} 
else {
  node = [[OCNode alloc] initWithOperator:ope value:0 fg:NULL fd:NULL];
	
	if ((position-Bimbric)==0) { /* argument absent */
		*error=3;
		return NULL;
	}
    
	tString=CopyPartialString(string,Bimbric,(position-1));
	
  [node setFg:ConstructTree(tString,(position-Bimbric),error)];
	
  free(tString);
	
  if ((length-position-1-Bimbric)==0) { /* argument absent */
    *error=3;
    return NULL;
  }
	
  tString=CopyPartialString(string,(position+1),(length-1-Bimbric));
	
  [node setFd:ConstructTree(tString,(length-position-1-Bimbric),error)];
	
  free(tString);
	
  return node;
}
}

/*************************************************************************/

int main(int argc, char ** argv) {
OCNode * test;
char tString[256];
char tString2[500];
short err=0;
#if MACOSX
NSAutoReleasePool * pool=[[NSAutoReleasePool alloc] init];
#endif

printf("Entrez votre expression parenthesee: \n");
scanf("%s",tString);
   
printf("%d\n",strlen(tString));

test=ConstructTree(tString,strlen(tString),&err);
printf("Type d'erreur: %d\n", err);
if (err==0) {
	printf("Evaluation: %f\n", [test ComputeTreewithXValue:1 YValue:1 error:&err]);
	printf("Type d'erreur: %d\n", err);
	tString2[0]='\0';
	[test WriteTree:tString2];
	printf("%s\n",tString);
	printf("%s\n",tString2);
}

#if MACOSX
[pool release];
exit(0);
#endif

return 0;
}
